unit Gr3D;

interface

uses
 SysUtils, Windows, Classes, Graphics, Forms, OpenGL;

const
 c1:real=0.8660;
 c2:real=0.5;
 nmax=40;
 nmax2:integer=20;
 nmax_sq=1521;
  {Basic 3D types}
type
 {Basic 3D function type}
 TArifmFunction=function(x,y:real):real;
 {Basic vector type}
 T_XYZ=record
  X:real;
  Y:real;
  Z:real
 end;
 {Polar vector type}
 TPolar=record
  hor:real;
  vert:real;
  rad:real
 end;
 {Cylinder Vector Type}
 TAxial=record
  A:real;
  Z:real;
  R:real
 end;
 {Special toroidal data}
 TToroid=record
  r1,r2:real;
  a1,a2:real
 end;
 {Screen output saving}
 TAngles=record
  HorAngle:Real;
  VertAngle:Real;
  DeltaAngle:Real;
 end;
 {Space Basis Type}
 TBasis=array[1..3] of T_XYZ;
 {Stack for saving current basis}
 TBasisStack=record
  BaseData:array[1..20] of TBasis;
  AngleData:array[1..20] of TAngles;
  Top:Word
 end;
 {Additional  type for quick rotation}
 TPolarBasis=array[1..3] of TPolar;
 {Basic 3D surface type}
 TGraph=array[1..nmax,1..nmax] of T_XYZ;
 TPolarGraph=array[1..nmax,1..nmax] of TPolar;
 TAxialGraph=array[1..nmax,1..nmax] of TAxial;
 TVertexes=array[1..nmax_sq,1..5] of T_XYZ;

var
 VertAngle,HorAngle:real; {Vertical and horizontal angles of rotation(measured in radians)}
 Basis,ExBasis:TBasis;
 GrData,GrDataEx:TGraph;
 GrForm:TForm;
 x2d0,y2d0:integer;
 Nul:T_XYZ;
 NulP:TPolar;
 NulA:TAxial;
 Light:T_XYZ;
 DeltaAngle:real;
 GraphicsReady:boolean;

 (*  Basic 3D procedures  *)
procedure LoadBasis;{Loads a basis depending on angles VertAngle and HorAngle}
procedure LoadBasisEx;{Loads external basis for lighted surfaces}
procedure PushBasis; {Saves the current basis to the Basis Stack}
procedure PopBasis; {Loads the basis from the stack}
procedure ResetBasis;
procedure DrawBasis(Scale:Word);{Draws the current basis}
procedure To2D(const R:T_XYZ; var x2d,y2d:integer);overload; {Translates 3D space to the screen}
procedure To2D(const R:T_XYZ; var Coords:TPoint);overload;
{Lighting mode}
procedure LoadLight(const lx,ly,lz:real);
{3D dot procedures}
procedure Pixel3D(const R:T_XYZ;color:TColor);overload; {3D Dot XYZ}
procedure Pixel3D(const R:TPolar;color:TColor);overload; {3D Dot Polar}
procedure Pixel3D(const R:TAxial;color:TColor);overload; {3D Dot Axial}
{3D line procedures}
procedure Line3D(const R1,R2:T_XYZ);overload; {3D line XYZ}
procedure Line3D(const R1,R2:TPolar);overload; {3D line Polar}
procedure Line3D(const R1,R2:TAxial);overload; {3D line Axial}
{3D polygon procedure}
procedure Polygon(const R1,R2,R3,R4:T_XYZ);
function GetColor(const R1,R2:T_XYZ):TColor;
 {Translation Procedures}
function PolarToXYZ(const R1:TPolar):T_XYZ; {TPolar -> T_XYZ}
function AxialToXYZ(const R1:TAxial):T_XYZ; {TAxial -> T_XYZ}
function ToroidToXYZ(const R:TToroid):T_XYZ;
procedure Radius(const R1,R2:TPolar);
procedure Evaluate(const f:TArifmFunction);
procedure EvaluatePolar(const f:TArifmFunction);
procedure EvaluateAxial(const f:TArifmFunction);
procedure EvaluateTor;
procedure Draw;
procedure DrawEx;{Draws the surface with lighting}
procedure SortVertex;
procedure ClearSpace; {Clears a current space}
 { ******** Basic Vector Operations *************** }
function Add(const R1,R2:T_XYZ):T_XYZ;
function Sub(const R1,R2:T_XYZ):T_XYZ;
function ScaMul(const R1,R2:T_XYZ):real;
function VecMul(const R1,R2:T_XYZ):T_XYZ;
function ConstMul(const R:T_XYZ;A:real):T_XYZ;
function MatrMul(const R:T_XYZ;M:TBasis):T_XYZ;
function Length(const R:T_XYZ):real;
 {********** Additional arithmetic functions ***********}
function Gr_Zero(x,y:real):real;

implementation

var
 c11,c12,c21,c22:real;{Simplified basis for 2D calculations}
 A11,A12,A13,
 A21,A22,A23:real;
 BasisStack:TBasisStack;
 Vertexes:TVertexes;

{ ********* Basic Vector Operations *********** }
function Add;
begin
 Add.X:=R1.X + R2.X;
 Add.Y:=R1.Y + R2.Y;
 Add.Z:=R1.Z + R2.Z
end;

function Sub;
begin
 Sub.X:=R2.X - R1.X;
 Sub.Y:=R2.Y - R1.Y;
 Sub.Z:=R2.Z - R1.Z
end;

function ScaMul;
begin
 ScaMul:=R1.X*R2.X + R1.Y*R2.Y + R1.Z*R2.Z
end;

function VecMul;
begin
 VecMul.X:=R1.Y*R2.Z - R2.Y*R1.Z;
 VecMul.Y:=R1.Z*R2.X - R1.X*R2.Z;
 VecMul.Z:=R1.X*R2.Y - R2.X*R1.Y
end;

function ConstMul;
begin
 ConstMul.X:=A*R.X;
 ConstMul.Y:=A*R.Y;
 ConstMul.Z:=A*R.Z;
end;

function MatrMul;
begin
 MatrMul.X:=ScaMul(R,M[1]);
 MatrMul.Y:=ScaMul(R,M[2]);
 MatrMul.Z:=ScaMul(R,M[3]);
end;

function Length;
var ab:real;
begin
 ab:=R.X*R.X + R.Y*R.Y + R.Z*R.Z;
 Length:=sqrt(ab);
end;

procedure LoadBasis;
var
 PBase:TPolarBasis;
begin
 {loading Polar Basis}
 PBase[1].hor:=HorAngle;
 PBase[3].hor:=HorAngle;
 PBase[1].vert:=-1*VertAngle;
 PBase[3].vert:=(pi/2) - VertAngle;
 PBase[1].rad:=1;
 PBase[3].rad:=1;
 {X ort}
 Basis[1]:=PolarToXYZ(PBase[1]);
 {Z ort}
 Basis[3]:=PolarToXYZ(PBase[3]);
 {Y ort}
 Basis[2]:=VecMul(Basis[3],Basis[1]);
 C11:=c1; C12:=-1*c1;
 C21:=c2; C22:=c2;
 A11:= C11*Basis[1].X + C12*Basis[2].X;
 A12:= C11*Basis[1].Y + C12*Basis[2].Y;
 A13:= Basis[1].Z + Basis[2].Z;
 A21:= C21*Basis[1].X + C22*Basis[2].X - Basis[3].X;
 A22:= C21*Basis[1].Y + C22*Basis[2].Y - Basis[3].Y;
 A23:= C21*Basis[1].Z + C22*Basis[2].Z - Basis[3].Z;
end;

procedure LoadBasisEx;
var
 PBase:TPolarBasis;
begin
 {loading Polar Basis}
 PBase[1].hor:=HorAngle;
 PBase[3].hor:=HorAngle;
 PBase[1].vert:=-1*VertAngle;
 PBase[3].vert:=(pi/2) - VertAngle;
 PBase[1].rad:=1;
 PBase[3].rad:=1;
 {X ort}
 ExBasis[1]:=PolarToXYZ(PBase[1]);
 {Z ort}
 ExBasis[3]:=PolarToXYZ(PBase[3]);
 {Y ort}
 ExBasis[2]:=VecMul(ExBasis[3],ExBasis[1]);
end;

procedure PushBasis;
begin
 BasisStack.BaseData[BasisStack.Top]:=Basis;
 BasisStack.AngleData[BasisStack.Top].HorAngle:=HorAngle;
 BasisStack.AngleData[BasisStack.Top].VertAngle:=VertAngle;
 BasisStack.AngleData[BasisStack.Top].DeltaAngle:=DeltaAngle;
 Inc(BasisStack.Top);
 HorAngle:=0;
 VertAngle:=0;
 DeltaAngle:=pi/10;
 LoadBasis;
end;

procedure PopBasis;
begin
 Dec(BasisStack.Top);
 Basis:=BasisStack.BaseData[BasisStack.Top];
 HorAngle:=BasisStack.AngleData[BasisStack.Top].HorAngle;
 VertAngle:=BasisStack.AngleData[BasisStack.Top].VertAngle;
 DeltaAngle:=BasisStack.AngleData[BasisStack.Top].DeltaAngle
end;

procedure ResetBasis;
begin
 HorAngle:=0;
 VertAngle:=0;
 DeltaAngle:=pi/10;
 LoadBasis
end;

procedure DrawBasis;
var
 SBase:TBasis;
 TC:TColor;
begin
 {Scaling X axis}
 SBase[1].X:=scale;
 SBase[2].X:=0;
 SBase[3].X:=0;
 {Scaling Y axis}
 SBase[1].Y:=0;
 SBase[2].Y:=Scale;
 SBase[3].Y:=0;
 {Scaling Z axis}
 SBase[1].Z:=0;
 SBase[2].Z:=0;
 SBase[3].Z:=Scale;
 {Drawing scaled basis}
 TC:=GrForm.Canvas.Pen.Color;
 GrForm.Canvas.Pen.Color:=rgb(255,0,0);
 Line3d(nul,SBase[1]);
 GrForm.Canvas.Pen.Color:=rgb(0,255,0);
 Line3d(nul,SBase[2]);
 GrForm.Canvas.Pen.Color:=rgb(0,0,255);
 Line3d(nul,SBase[3]);
 GrForm.Canvas.Pen.Color:=TC;
end;

{Translation 3D->2D procedure}
procedure To2D(const R:T_XYZ;var x2d,y2d:integer);overload;
begin
 with R do
 begin
  x2d:=x2d0+round(A11*X + A12*Y + A13*Z);
  y2d:=y2d0+round(A21*X + A22*Y + A23*Z)
 end;
end;

procedure To2D(const R:T_XYZ;var Coords:TPoint);overload;
begin
 Coords.X:=x2d0 + round(A11*R.X + A12*R.Y + A13*R.Z);
 Coords.Y:=y2d0 + round(A21*R.X + A22*R.Y + A23*R.Z)
end;

procedure LoadLight;
begin
 Light.X:=127*lx/sqrt(lx*lx+ly*ly+lz*lz);
 Light.Y:=127*ly/sqrt(lx*lx+ly*ly+lz*lz);
 Light.Z:=127*lz/sqrt(lx*lx+ly*ly+lz*lz);
end;

{ ********** 3D Dots procedures ************ }
procedure Pixel3D(const R:T_XYZ;Color:TColor);overload;
var
 x2d,y2d:integer;
begin
 with R do
 begin
  x2d:=x2d0+round(A11*X + A12*Y + A13*Z);
  y2d:=y2d0+round(A21*X + A22*Y + A23*Z)
 end;
 GrForm.Canvas.Pixels[x2d,y2d]:=color;
end;

procedure Pixel3D(const R:TAxial;color:TColor);overload;
var
 TR:T_XYZ;
begin
 TR:=AxialToXYZ(R);
 Pixel3D(TR,Color);
end;

procedure Pixel3D(const R:TPolar;color:TColor);overload;
var
 TR:T_XYZ;
begin
 TR:=PolarToXYZ(R);
 Pixel3D(TR,Color);
end;

{ ********** 3D Lines ************** }
procedure line3d(const R1,R2:T_XYZ);overload;
var
 x2d1,y2d1,x2d2,y2d2:integer;
begin
 x2d1:=x2d0+round(A11*R1.X + A12*R1.Y + A13*R1.Z);
 y2d1:=y2d0+round(A21*R1.X + A22*R1.Y + A23*R1.Z);
 x2d2:=x2d0+round(A11*R2.X + A12*R2.Y + A13*R2.Z);
 y2d2:=y2d0+round(A21*R2.X + A22*R2.Y + A23*R2.Z);
 GrForm.Canvas.moveto(x2d1,y2d1);
 GrForm.Canvas.lineto(x2d2,y2d2);
end;

procedure Line3D(const R1,R2:TPolar);overload;
var
 TR1,TR2:T_XYZ;
begin
 TR1:=PolarToXYZ(R1);
 TR2:=PolarToXYZ(R2);
 Line3D(TR1,TR2);
end;

procedure Line3D(const R1,R2:TAxial);overload;
var
 TR1,TR2:T_XYZ;
begin
 TR1:=AxialToXYZ(R1);
 TR2:=AxialToXYZ(R2);
 Line3D(TR1,TR2);
end;

{ ******** Polygon procedure ****************** }
procedure Polygon;
var
 P:array[1..4] of TPoint;
begin
 To2D(R1,P[1]);
 To2D(R2,P[2]);
 To2D(R3,P[3]);
 To2D(R4,P[4]);
 GrForm.Canvas.Polygon(P);
end;

function GetColor;
var
 CR,CR1:T_XYZ;
 L,Col:real;
 Intens:byte;
begin
 CR:=VecMul(R1,R2);
 L:=1/Length(CR);
 CR1:=ConstMul(CR,L);
 Col:=ScaMul(CR1,Light);
 Col:=abs(Col);
 if Col<127 then
  Intens:=127+round(Col)
 else
  Intens:=255;
 GetColor:=rgb(Intens,Intens,Intens);
end;

{ ************* Translation procedures ************ }
function PolarToXYZ;
begin
 PolarToXYZ.Z:=R1.rad*sin(R1.vert);
 PolarToXYZ.X:=R1.rad*cos(R1.vert)*cos(R1.hor);
 PolarToXYZ.Y:=R1.rad*cos(R1.vert)*sin(R1.hor)
end;

function AxialToXYZ;
begin
 AxialToXYZ.Z:=R1.Z;
 AxialToXYZ.X:=R1.R*cos(R1.A);
 AxialToXYZ.Y:=R1.R*sin(R1.A)
end;

function ToroidToXYZ;
begin
 ToroidToXYZ.Z:=R.r2*sin(R.a2);
 ToroidToXYZ.X:=(R.r1 + R.r2*cos(R.a2))*cos(R.a1);
 ToroidToXYZ.Y:=(R.r1 + R.r2*cos(R.a2))*sin(R.a1)
end;
 
procedure Radius;
var
 XYZ1,XYZ2:T_XYZ;
begin
 XYZ1:=PolarToXYZ(R1);
 XYZ2:=PolarToXYZ(R2);
 line3d(XYZ1,XYZ2);
end;

procedure evaluate;
var
 i,j:integer;
begin
 for i:=1 to nmax do
  for j:=1 to nmax do
   with GrData[i,j] do
   begin
    X:=10*(i-nmax2);
    Y:=10*(j-nmax2);
    Z:=f(X,Y)
   end;
end;

procedure evaluatePolar;
var
 i,j:integer;
 cvert,chor:real;
 RT:TPolar;
begin
 cvert:=pi/nmax;
 chor:=2*pi/(nmax-1);
 for i:=1 to nmax do
  for j:=1 to nmax do
  begin
   RT.hor:=chor*(i-nmax2);
   RT.vert:=cvert*(j-nmax2);
   RT.rad:=f(RT.hor,RT.vert);
   GrData[i,j]:=PolarToXYZ(RT);
  end
end;

procedure evaluateAxial;
var
 i,j:integer;
 cA:real;
 RA:TAxial;
begin
 cA:=2*pi/(nmax-1);
 for i:=1 to nmax do
  for j:=1 to nmax do
  begin
   RA.A:=cA*i;
   RA.R:=5*j;
   RA.Z:=f(RA.R,RA.A);
   GrData[i,j]:=AxialToXYZ(RA);
  end
end;

procedure EvaluateTor;
var
 i,j:integer;
 cA,cB:real;
 RA:TToroid;
begin
 cA:=2*pi/(nmax-1);
 cB:=2*pi/(nmax-1);
 for i:=1 to nmax do
  for j:=1 to nmax do
  begin
   RA.a1:=cA*i;
   RA.a2:=cB*(j-nmax2);
   RA.r1:=100; Ra.r2:=50;
   GrData[i,j]:=ToroidToXYZ(RA);
  end
end;

procedure draw;
var
 R,R1:T_XYZ;
 i,j:integer;
begin
 ClearSpace;
 GrForm.Canvas.Pen.Color:=rgb(63,255,127);
 R1:=GrData[1,1];
 for i:=1 to nmax do
 begin
  for j:=1 to nmax do
  begin
   R:=GrData[i,j];
   line3d(R1,R);
   R1:=R;
  end;
  R1:=GrData[i+1,1];
 end;
 R1:=GrData[1,1];
 for j:=1 to nmax do
 begin
  for i:=1 to nmax do
  begin
   R:=GrData[i,j];
   line3d(R1,R);
   R1:=R;
  end;
  R1:=GrData[1,j+1];
 end;
end;

procedure SortVertex;
var
 i,j,k,m:word;
 S:real;
 Max:real;
 Temp:array[1..5] of T_XYZ;
begin
 k:=1;
 for i:=1 to nmax-1 do
  for j:=1 to nmax-1 do
  begin
   Vertexes[k,1]:=GrDataEx[i,j];
   Vertexes[k,2]:=GrDataEx[i+1,j];
   Vertexes[k,3]:=GrDataEx[i+1,j+1];
   Vertexes[k,4]:=GrDataEx[i,j+1];
   Vertexes[k,5]:=Add(Add(Vertexes[k,1],Vertexes[k,2]),Add(Vertexes[k,3],Vertexes[k,4]));
   inc(k);
  end;
  for i:=1 to nmax_sq do
  begin
   Max:=-100000000;
   for j:=i to k do
   begin
    S:=Vertexes[j,5].X + Vertexes[j,5].Y + Vertexes[j,5].Z;
    if S>=Max then
    begin
     Max:=S;
     m:=j;
    end;
   end;
   for j:=1 to 5 do
    Temp[j]:=Vertexes[i,j];
   for j:=1 to 5 do
    Vertexes[i,j]:=Vertexes[m,j];
   for j:=1 to 5 do
    Vertexes[m,j]:=Temp[j];
  end;
end;

procedure DrawEx;
var
 i,j:word;
 TCB,TCP:TColor;
 PolyColor:TColor;
 DR1,DR2:T_XYZ;
begin
 PushBasis;
 TCB:=grForm.Canvas.Brush.Color;
 TCP:=grForm.Canvas.Pen.Color;
 for i:=1 to nmax do
  for j:=1 to nmax do
   GrDataEx[i,j]:=MatrMul(GrData[i,j],ExBasis);
 SortVertex;
 for i:=nmax_sq downto 1 do
  begin
   DR1:=Sub(Vertexes[i,2],Vertexes[i,1]);
   DR2:=Sub(Vertexes[i,4],Vertexes[i,1]);
   PolyColor:=GetColor(DR1,DR2);
   grForm.Canvas.Brush.Color:=PolyColor;
   grForm.Canvas.Pen.Color:=PolyColor;
   Polygon(Vertexes[i,1],Vertexes[i,2],Vertexes[i,3],Vertexes[i,4]);
  end;
 grForm.Canvas.Brush.Color:=TCB;
 grForm.Canvas.Pen.Color:=TCP;
 PopBasis;
end;

procedure ClearSpace;
var
 FormRect:TRect;
 TC:TColor;
begin
 TC:=GrForm.Canvas.Brush.Color;
 GrForm.Canvas.Brush.Color:=clBlack;
 FormRect:=Rect(0,0,GrForm.Width,GrForm.Height);
 GrForm.Canvas.FillRect(FormRect);
 GrForm.Canvas.Brush.Color:=TC;
end;

function Gr_Zero;
begin
 Gr_Zero:=0;
end;

initialization

begin
 VertAngle:=0;
 HorAngle:=0;
 DeltaAngle:=pi/10;
 LoadBasis;
 Nul.X:=0;
 Nul.Y:=0;
 Nul.Z:=0;
 NulP.hor:=0;
 NulP.vert:=0;
 NulP.rad:=0;
 NulA.A:=0;
 NulA.Z:=0;
 NulA.R:=0;
 LoadLight(1,1,1);
 Evaluate(Gr_Zero);
end;

end.
